import chess.pgn
import bunseki.util as util
import bunseki.asklichess as ali
from collections import defaultdict 
import argparse

parser = argparse.ArgumentParser() 
parser.add_argument('-w','--white',dest='COLOR', help= 'is the repertoire for white', type=int, choices=[0,1], required=True)
parser.add_argument('-i','--pgnfile',dest='PGNFILE', help= 'path to pgn file to parse', type=str, required=True)
parser.add_argument('-m','--minmoves',dest='MINMOV', help= 'suggestions musst have been played this many times', type=int, default = 500)
parser.add_argument('-u','--utilitycuy',dest='UTILITYCUT', help= 'cutoff based on utility..i.e. probability that the top move is on board', type=float, default = 0)

parser.add_argument('-d','--database',dest='DATABASE', help= '0 = master database 1 = lichess', type=int, default = 0)
parser.add_argument('-s','--lichess_strength',dest='STRENGTH', help= 'select playing strength for lichess db', type=int,nargs='+', default = [1600,1800])
parser.add_argument('-t','--lichess_format',dest='TIMECTL', help= 'select time control for lichess db', type=str,nargs='+', default = ['blitz','rapid'])


parser.add_argument('-a','--showallnodes',dest='PRINTALL', help= 'print all to nodes even if they pass the filter',action = "store_true")
parser.add_argument('-p','--minpercentage',dest='MINPERC', help= 'suggestions musst have been played this many times', type=int, default = 3)
parser.add_argument('-c','--coverage',dest='MINCOV', help= 'replies musst cover this percentage of moves', type=int, default = 100)
parser.add_argument('--maxparse',dest='MAXPARSE', help= 'my studies have a few outdated chapters that i put at the end', type=int, default = 100)


def proba_calculation(gn, moves, movesum):

    movcntdi = { m['san']:ali.sumdi(m)  for m in moves}
    for child in gn.variations: #  i could calculate the percentage of games that end up here...
        if movesum == 0: 
            child.proba = 0
        else:
            child.proba = movcntdi.get(child.san(),0)/ movesum


def main(): 
    args = parser.parse_args()
    db = ali.lichess(args)

    ####
    # collect the game in a huge tree
    # we could also just make a list of fens, but the tree makes it easy 
    # to cut a whole branch when there are too few games
    ####
    repo = open(args.PGNFILE)
    games = []
    while True:
        stuff = chess.pgn.read_game(repo)
        if not stuff or len(games)==args.MAXPARSE:
            break
        games.append(stuff)
    game = util.merge(games)

    fen = game.board().fen()
    moves, total_games, openingname = db.ask(fen)
    game.proba = 1
    game.openingname = openingname
    if args.COLOR ==0:
        proba_calculation(game, moves, total_games)

    ###############
    # for fens get the game-node and lichess-db answer
    ##############
    fens = defaultdict(list)
    def visit(gn):
        if gn.ply() % 2  == args.COLOR:  # do we need to analyze this?
            gn.proba = 1 
            enemy_moves = [str(v.san()) for v in  gn.variations] # UCI has redundant codes :(
            if  enemy_moves:
                # lets do a lookup 
                fen = gn.board().fen()
                if fen in fens:
                    moves,movesum = fens[fen][0][:2]

                moves, movesum, opening= db.ask(fen)

                # this blocck is for proba calculation
                proba_calculation(gn,moves,movesum) 
                gn.opening = opening

                fens[fen].append([moves,movesum,gn]) 
                if movesum < args.MINMOV*2:
                    return

        for child in gn.variations:
            visit(child)
    for g in game.variations:
        visit(g)

    db.end()

    #########################
    # so, now we can evaluate what we have... 
    ##########################

    # sort by node weight
    #keys = list(fens.keys())
    #keys.sort(key = lambda x: fens[x][0][1], reverse = True)



    bunsekitachi = []

    for k in fens.keys(): 
        mo_s_gn_li  = fens[k]
        # all the moves from all game nodes that share the fen
        enemy_moves = [str(v.san()) for mosgn in mo_s_gn_li  for v in  mosgn[2].variations]
        res,perc_miss = ali.analyse2(mo_s_gn_li[0][0],enemy_moves, 
                              minmov=args.MINMOV, 
                              target_share  = args.MINCOV/100,
                              minperc= args.MINPERC/100)
        bunsekitachi.append(bunseki( [e[2] for e in mo_s_gn_li ]  ,args.COLOR, res, perc_miss   ))

    bunsekitachi.sort(key = lambda x: x.frequency*x.perc_miss)

    for e in bunsekitachi: 
        if (e.perc_miss > 0 and e.utility > args.UTILITYCUT)  or args.PRINTALL:
            e.print()
        




class bunseki():
    def __init__(self,gn,color,res_str,perc_miss):
        self.gn = gn
        self.color= color
        self.moves=res_str
        self.perc_miss = perc_miss
        self.board = gn[0].board()
        self.frequency = sum(map( util.getimportance, self.gn))*100
        self.utility = self.frequency * self.perc_miss

    def print(self):
        print(f"possition frequency:{ self.frequency:.2f} utility:{self.utility:.2f}")
        op = self.gn[0].opening
        if op:
            print (f"{op['eco']} {op['name']} ")
        print ()


        if self.color ==1:
            print(self.board.unicode(empty_square=' ',invert_color=True))
        else:
            print(self.board.transform(chess.flip_vertical).transform(chess.flip_horizontal).unicode(empty_square=' ',invert_color=True))

        print (self.moves)
        print ()
            
        for node in self.gn:
            print(util.pgn(node))
        print(self.board.fen())
        print(f"https://lichess.org/analysis/{self.board.fen().replace(' ','_')}")
        #print( self.gn.get('opening',''))
        print("\n\n\n")

if __name__ == '__main__':
    main()



# based on https://github.com/permutationlock/merge-pgn
import chess.pgn
import sys
import chess
import os 
import pickle
def merge(games):
    master_node = chess.pgn.Game()

    mlist = []
    for game in games:
        mlist.extend(game.variations)

    variations = [(master_node, mlist)]
    done = False

    while not done:
        newvars = []
        done = True
        for vnode, nodes in variations:
            newmoves = {}
            for node in nodes:
                if node.move is None:
                    continue
                elif node.move not in list(newmoves):
                    nvnode = vnode.add_variation(node.move)
                    if len(node.variations) > 0:
                        done = False
                    newvars.append((nvnode, node.variations))
                    newmoves[node.move] = len(newvars) - 1
                else:
                    nvnode, nlist = newvars[newmoves[node.move]]
                    if len(node.variations) > 0:
                        done = False
                    nlist.extend(node.variations)
                    newvars[newmoves[node.move]] = (nvnode, nlist)
        variations = newvars

    return master_node

def getimportance(node):

    res = 1
    values =[] 
    while node.parent:
        res*= node.proba
        values.append(node.proba)
        node = node.parent
    values.reverse()
    #print("debug_proba:", values)
    return res


def history(node):
    
    r=[]
    while node.parent:
        r.append(str(node.move))
        node = node.parent
    r.reverse()
    return r

def pgn(node):

    g = chess.pgn.Game()
    for move in history(node):
        g= g.add_variation(chess.Move.from_uci(move))

    exporter = chess.pgn.StringExporter(headers=False, variations=True, comments=True)
    return g.game().accept(exporter)



class cacher():
    def __init__(self,cachename): 
        self.cachename = f".{cachename}"
        if os.path.exists(self.cachename): 
            self.cache = pickle.load( open( self.cachename, "rb" ) )
        else:
            self.cache={}

    def write(self):
        pickle.dump( self.cache, open( self.cachename, "wb" ) )

    def call(self,f, key): 
        if key in self.cache:
            return self.cache[key]
        else:
            r = f()
            self.cache[key] = r 
        return r




# we interact with this: https://github.com/niklasf/lila-openingexplorer



import requests as req
from pprint import pprint
import time

session = req.Session()

def ask(fen, args):
    print(".",end='')
    if args.DATABASE ==1:
        parm = {'fen':fen, 
                'topGames': 0,
                'recentGames':0, 
                'moves':'15',
                'variant':"standard",
                'speeds[]':args.TIMECTL,
                'ratings[]':args.STRENGTH}
        res = session.get('https://explorer.lichess.ovh/lichess', params=parm)
    else:
        parm = {'fen':fen, 'topGames': 0, 'moves':'15' }
        res = session.get('https://explorer.lichess.ovh/master', params=parm)

    if  res.status_code == 200:
        js = res.json()
        time.sleep(.5)
        return js['moves'], sumdi(js) 
    else:
        print (res.status_code)
        print(res.text)
        assert False



def sumdi(di):
    return di['black']+di['white']+di['draws']


def fmt_stst(my,al):
    return int((my/al)*100)
def fmt_q(di):
    a= sumdi(di)
    i = lambda x: int(x*100)
    return [ i(di[v]/a) for v in['white','draws','black']   ]

def analyse(lidi,okmoves,minmov,minperc, games_total):
    ret = []
    dirty= False
    sum_moves = sum([ sumdi(di) for di in lidi   ] )

    cum_mv =0 
    aru = 0
    for di in lidi: 
        if cum_mv < minperc*sum_moves: # havent reached 90% yet
            mvcnt = sumdi(di)
            if mvcnt > minmov: # more than minmov played
                cum_mv+= mvcnt 
                freq = f"freq: { (mvcnt*100)/games_total :.2f}"
                if  di['san'] not in okmoves:
                    ret.append( f"\t{di['san']}\t{mvcnt}({fmt_stst(mvcnt,sum_moves)})\t!! {fmt_q(di)}"   )
                    dirty = True
                else:
                    aru+=mvcnt
                    ret.append( f"\t{di['san']}\t{mvcnt}({fmt_stst(mvcnt,sum_moves)})\tOK {fmt_q(di)}"   )
    
    
    if not dirty:
        ret = ''
    else:
        ret = '\n'.join(ret)
    return ret, minperc*sum_moves- aru




#!/usr/bin/env python
from __future__ import absolute_import
from __future__ import print_function

import os
import sys
import subprocess
import re

from setuptools import setup
from setuptools.command.sdist import sdist as _sdist
from setuptools.command.install import install as _install

VERSION_PY = """
# This file is originally generated from Git information by running 'setup.py
# version'. Distribution tarballs contain a pre-generated copy of this file.
__version__ = '%s'
"""


def update_version_py():
    if not os.path.isdir(".git"):
        print("This does not appear to be a Git repository.")
        return
    try:
        #p = subprocess.Popen(["git", "describe","--tags", "--always"],
        #        stdout=subprocess.PIPE)
        p = subprocess.Popen("git rev-list HEAD --count".split(),
                stdout=subprocess.PIPE)


    except EnvironmentError:
        print("unable to run git, leaving bunseki/_version.py alone")
        return
    stdout = p.communicate()[0]
    if p.returncode != 0:
        print("unable to run git, leavi bunseki/_version.py alone")
        return
    ver = "0.0."+stdout.decode('utf-8').strip()
    
    #ver = str(int(ver,16)) # pypi doesnt like base 16
    f = open("bunseki/_version.py", "w")
    f.write(VERSION_PY % ver)
    f.close()
    print("set bunseki/_version.py to '%s'" % ver)


def get_version():
    try:
        f = open("bunseki/_version.py")
    except EnvironmentError:
        return None
    for line in f.readlines():
        mo = re.match("__version__ = '([^']+)'", line)
        if mo:
            ver = mo.group(1)
            return ver
    return None


class sdist(_sdist):

    def run(self):
        update_version_py()
        self.distribution.metadata.version = get_version()
        return _sdist.run(self)


class install(_install):

    def run(self):
        _install.run(self)

    def checkProgramIsInstalled(self, program, args, where_to_download,
                                affected_tools):
        try:
            subprocess.Popen([program, args], stderr=subprocess.PIPE, stdout=subprocess.PIPE)
            return True
        except EnvironmentError:
            # handle file not found error.
            # the config file is installed in:
            msg = "\n**{0} not found. This " \
                  "program is needed for the following "\
                  "tools to work properly:\n"\
                  " {1}\n"\
                  "{0} can be downloaded from here:\n " \
                  " {2}\n".format(program, affected_tools,
                                  where_to_download)
            sys.stderr.write(msg)

        except Exception as e:
            sys.stderr.write("Error: {}".format(e))

setup(
    name='bunseki',
    version=get_version(),
    author='Stefan Mautner',
    author_email='myl4stn4m3@cs.uni-freiburg.de',
    packages=[ 'bunseki' ],
    scripts=[ ],
    include_package_data=True,
    package_data={},
    url='https://github.com/smautner/',
    license='LICENSE',
    description='compares moves in a chess repertoire against lichess database',
    install_requires=[
        'python-chess','requests'
    ],
    entry_points = { 'console_scripts': 
                    [
                        'bunseki=bunseki.analyze:main',
                        'bunseki_repobuilder=bunseki.repobuilder:main'
                    ] },
    cmdclass={'sdist': sdist, 'install': install}
)
